#include <sqlextract/filter/arbval.h>
#include <sqlextract/util.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

struct arbval_t *arbval_create_null(void) {
	struct arbval_t *av = lf_calloc(1, sizeof(struct arbval_t));
	if (av == NULL)
		return NULL;
	av->type = ARBVAL_TYPE_NULL;
	return av;
}
struct arbval_t *arbval_create_integer(long value) {
	struct arbval_t *av = lf_calloc(1, sizeof(struct arbval_t));
	if (av == NULL)
		return NULL;
	av->type = ARBVAL_TYPE_INTEGER;
	av->value.integer = value;
	return av;
}
struct arbval_t *arbval_create_string(const char *value) {
	struct arbval_t *av = lf_calloc(1, sizeof(struct arbval_t));
	if (av == NULL)
		return NULL;
	av->type = ARBVAL_TYPE_STRING;
	av->value.string = lf_strdup(value);
	if (av->value.string == NULL) {
		free(av);
		return NULL;
	}
	return av;
}
struct arbval_t *arbval_create_real(double value) {
	struct arbval_t *av = lf_calloc(1, sizeof(struct arbval_t));
	if (av == NULL)
		return NULL;
	av->type = ARBVAL_TYPE_REAL;
	av->value.real = value;
	return av;
}

void arbval_destroy(struct arbval_t *av) {
	if(av==NULL)
		return;
	arbval_typecast_null(av); // Make sure storage is cleared
	free(av);
}

struct arbval_t *arbval_clone(struct arbval_t *av) {
	if (av == NULL)
		return NULL;
	switch (av->type) {
	case ARBVAL_TYPE_NULL:
		return arbval_create_null();
	case ARBVAL_TYPE_INTEGER:
		return arbval_create_integer(av->value.integer);
	case ARBVAL_TYPE_STRING:
		return arbval_create_string(av->value.string);
	case ARBVAL_TYPE_REAL:
		return arbval_create_real(av->value.real);
	}
	return NULL;
}

void arbval_clone_into(struct arbval_t *dst, struct arbval_t *src) {
	arbval_typecast_null(dst); // Make sure storage is cleared
	dst->type = src->type;
	switch (src->type) {
	case ARBVAL_TYPE_NULL:
		dst->value.integer = 0;
		return;
	case ARBVAL_TYPE_INTEGER:
		dst->value.integer = src->value.integer;
		return;
	case ARBVAL_TYPE_STRING:
		dst->value.string = lf_strdup(src->value.string);
		return;
	case ARBVAL_TYPE_REAL:
		dst->value.real = src->value.real;
		return;
	}
}

void arbval_typecast_null(struct arbval_t *av) {
	switch (av->type) {
	case ARBVAL_TYPE_NULL:
		av->value.integer = 0;
		break;
	case ARBVAL_TYPE_INTEGER:
		av->value.integer = 0;
		break;
	case ARBVAL_TYPE_STRING:
		free(av->value.string);
		av->value.integer = 0;
		break;
	case ARBVAL_TYPE_REAL:
		av->value.real = 0;
		break;
	}
	av->type = ARBVAL_TYPE_NULL;
}

void arbval_typecast_integer(struct arbval_t *av) {
	long tmp_long;
	switch (av->type) {
	case ARBVAL_TYPE_NULL:
		av->value.integer = 0;
		break;
	case ARBVAL_TYPE_INTEGER:
		break;
	case ARBVAL_TYPE_STRING:
		tmp_long = atol(av->value.string);
		free(av->value.string);
		av->value.integer = tmp_long;
		break;
	case ARBVAL_TYPE_REAL:
		tmp_long = av->value.real;
		av->value.integer = tmp_long;
		break;
	}
	av->type = ARBVAL_TYPE_INTEGER;
}

void arbval_typecast_string(struct arbval_t *av) {
	switch (av->type) {
	case ARBVAL_TYPE_NULL:
		av->value.string = lf_strdup("");
		break;
	case ARBVAL_TYPE_INTEGER:
		if (asprintf(&av->value.string, "%ld", av->value.integer) == -1) {
			av->value.string = NULL;
		}
		break;
	case ARBVAL_TYPE_STRING:
		break;
	case ARBVAL_TYPE_REAL:
		if (asprintf(&av->value.string, "%g", av->value.real) == -1) {
			av->value.string = NULL;
		}
		break;
	}
	av->type = av->value.string == NULL ? ARBVAL_TYPE_NULL : ARBVAL_TYPE_STRING;
}

void arbval_typecast_real(struct arbval_t *av) {
	double tmp_double;
	switch (av->type) {
	case ARBVAL_TYPE_NULL:
		av->value.real = 0;
		break;
	case ARBVAL_TYPE_INTEGER:
		tmp_double = av->value.integer;
		av->value.real = tmp_double;
		break;
	case ARBVAL_TYPE_STRING:
		tmp_double = atof(av->value.string);
		free(av->value.string);
		av->value.real = tmp_double;
		break;
	case ARBVAL_TYPE_REAL:
		break;
	}
	av->type = ARBVAL_TYPE_REAL;
}

int arbval_is_true(struct arbval_t *av) {
	switch (av->type) {
	case ARBVAL_TYPE_NULL:
		return 0;
	case ARBVAL_TYPE_INTEGER:
		return av->value.integer != 0;
	case ARBVAL_TYPE_STRING:
		return av->value.string[0] != '\0';
	case ARBVAL_TYPE_REAL:
		return !!av->value.real;
	}
	return 0;
}

int arbval_is_false(struct arbval_t *av) {
	return !arbval_is_true(av);
}

